@@ -1,6 +1,5 @@
-import json
 from datetime import datetime
-from typing import Callable, Dict, Any, List
+from typing import Callable, Dict, Any, List, Generator, Union
 
 import aiohttp
 from typeguard import typechecked
@@ -139,6 +138,81 @@ class Completion:
             logging.warning(f"No cost per 1000 tokens found for model {model_name}, using default value of 0.0")
             return 0.0
 
+    def stream_create(self,
+               prompt: str,
+               prompt_version: str,
+               prompt_version_description: str,
+               preprocess_prompt: Callable = _preprocess_prompt,
+               process_response: Callable = _process_response,
+               preprocess_prompt_args: Dict[Any, Any] = None,
+               process_response_args: Dict[Any, Any] = None) -> Union[Dict[Any, Any], Generator[Any, Any, Any]]:
+        """
+        Create a completion request and store it in MongoDB
+
+        Args:
+            prompt: The prompt to use for the completion
+            prompt_version: The version of the prompt (e.g. "1.0")
+            prompt_version_description: A description of the prompt version (e.g. "Added more diverse examples")
+            preprocess_prompt: A function that preprocesses the prompt
+            process_response: A function that processes the response
+            preprocess_prompt_args: A dictionary of arguments to pass to the preprocess_prompt function
+            process_response_args: A dictionary of arguments to pass to the process_response function
+
+        Returns:
+            A generator containing the completion responses, token by token
+        """
+        self.model_params['stream'] = True
+
+        if preprocess_prompt_args is None:
+            preprocess_prompt_args = {}
+        if process_response_args is None:
+            process_response_args = {}
+
+        preprocessed_prompt = preprocess_prompt(prompt, **preprocess_prompt_args)
+        response = self.openai_client.Completion.create(**self.model_params, prompt=preprocessed_prompt)
+
+        all_logprobs = {'tokens': [], 'token_logprobs': [], 'top_logprobs': [], 'text_offset': []}
+        completion_text = ""
+        token_number = 0
+
+        for resp in response:                
+            logprobs = resp['choices'][0]['logprobs']                
+            all_logprobs['tokens'].extend(logprobs['tokens'])
+            all_logprobs['token_logprobs'].extend(logprobs['token_logprobs'])
+            all_logprobs['top_logprobs'].extend(logprobs['top_logprobs'])
+            all_logprobs['text_offset'].extend(logprobs['text_offset'])
+            
+            completion_text += resp['choices'][0]['text']  
+            token_number += 1
+            yield resp
+        
+        openai_response = resp.__dict__['_previous']
+        openai_response['choices'][0]['logprobs'] = all_logprobs
+        openai_response['choices'][0]['text'] = completion_text
+
+        request_costs = self._calculate_request_costs(token_number)
+
+        document = {
+            "prompt": preprocessed_prompt,
+            "prompt_version": prompt_version,
+            "prompt_version_description": prompt_version_description,
+            "completion_text":  completion_text,
+            "openai_response": openai_response,
+            "model_params": self.model_params,
+            "created_at": str(datetime.utcnow()),
+            "experiment_metadata": self.experiment_metadata,
+            "project_metadata": self.project_metadata,
+            "request_cost_in_usd": request_costs
+        }
+
+        if self.extra_params is not None:
+            document.update(self.extra_params)
+
+        mongo_db = self.mongo_client[self.config.mongo_db_name]
+        mongo_collection = mongo_db[self.use_case_name]
+        mongo_collection.insert_one(document)
+
+    
     def create(self,
                prompt: str,
                prompt_version: str,
@@ -198,6 +272,7 @@ class Completion:
 
         return document
 
+
     async def async_create(self,
                            prompt: str,
                            prompt_version: str,
@@ -328,3 +403,54 @@ class Completion:
         """
         mongo_db = config.mongo_client[config.mongo_db_name]
         return list(mongo_db.list_collection_names())
+    
+    def save_to_mongo(self,
+               prompt: str,
+               response: str,
+               prompt_version: str,
+               prompt_version_description: str,
+               process_response: Callable = _process_response,
+               process_response_args: Dict[Any, Any] = None) -> Dict[Any, Any]:
+        """
+        Save the given prompt and OpenAI response to MongoDB. This functions is for use cases where you want to save the prompt and response without calling the OpenAI API.
+
+        Args:
+            prompt: The prompt submitted to OpenAI
+            response: The response object OpenAI API returned as completion
+            prompt_version: The version of the prompt (e.g. "1.0")
+            prompt_version_description: A description of the prompt version (e.g. "Added more diverse examples")
+            process_response: A function that processes the response
+            preprocess_prompt_args: A dictionary of arguments to pass to the preprocess_prompt function
+            process_response_args: A dictionary of arguments to pass to the process_response function
+
+        Returns:
+            A dictionary containing the completion request and response
+        """
+
+     
+        if process_response is not None:
+            response = process_response(response, **process_response_args)
+
+        request_costs = self._calculate_request_costs(response["usage"]["total_tokens"])
+
+        document = {
+            "prompt": prompt,
+            "prompt_version": prompt_version,
+            "prompt_version_description": prompt_version_description,
+            "completion_text": response["choices"][0]["text"],
+            "openai_response": response.__dict__['_previous'],
+            "model_params": self.model_params,
+            "created_at": str(datetime.utcnow()),
+            "experiment_metadata": self.experiment_metadata,
+            "project_metadata": self.project_metadata,
+            "request_cost_in_usd": request_costs
+        }
+
+        if self.extra_params is not None:
+            document.update(self.extra_params)
+
+        mongo_db = self.mongo_client[self.config.mongo_db_name]
+        mongo_collection = mongo_db[self.use_case_name]
+        mongo_collection.insert_one(document)
+
+        return document
